/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "frame_loc.hh"
#include "gpsdate.hh"
#include "driver_table.hh"
#include <sys/stat.h>
#include <stdexcept>
#include <unistd.h>
#include <iostream>
#include <algorithm>

using namespace sends;
using namespace std;

//======================================  Constructor
frame_loc::frame_loc(const std::string& location, const std::string& prefix, 
		     gps_type start, gps_type stop, gps_type dT, 
		     driver_type drive)
  : mLocation(location), mPrefix(prefix), mStart(start), mStop(stop), 
    mDuration(dT), mDriver(drive)
{}

//======================================  Destructor
frame_loc::~frame_loc(void) {
}

//======================================  Dump a location entry
std::ostream& 
frame_loc::dump(std::ostream& out) const {
    out << "    " << mPrefix << "  " << mStart 
	<< "  " << mStop << "  " << mDuration 
	<< "  " << mDriver << "  " << mLocation << endl;
    return out;
}

//======================================  Put high-order digits
static int
putstr(string& str, gps_type gps, int nDigit) {
    int nd = 0;
    if (gps < 10) {
	str += '0' + gps;
    } else {
	nd =putstr(str, gps/10, nDigit);
	if (!nDigit || nd < nDigit) {
	    str += '0' + (gps % 10);
	}
    }
    return nd+1;
}

//======================================  Put digits with optional padding
static void
padstr(string& str, gps_type gps, int nDigit) {
    if (gps > 9 || nDigit > 1) padstr(str, gps/10, nDigit-1);
    str += '0' + (gps % 10);
}

//======================================  Substitute gps into path
string
frame_loc::file_path(gps_type t, bool limit) const {
    string str;
    if (limit && !in_range(t)) return str;
    str.reserve(128);

    //----------------------------------  Get the frame start time.
    gps_type t0 = align_gps(t);

    //----------------------------------  Construct the frame path string
    const char* fmt = mLocation.c_str();
    while (*fmt) {
	char fchr = *fmt++;
	if (fchr != '%') {
	    str += fchr;
	} else {
	    //bool zpad = false;
	    //if (*fmt == '0') {
	    //zpad = true;
	    //*fmt++;
	    //}
	    int  cnt=0;
	    while (*fmt >= '0' && *fmt <= '9') {
		cnt *= 10;
		cnt += *fmt++ - '0';
	    }
	    gps_date gdate(t0);
	    switch (*fmt++) {
	    case 'd': //  GPS digits
		putstr(str, t0, cnt);
		break;
	    case 'D': //  Day of year
		padstr(str, gdate.day_in_year(), cnt);
		break;
	    case 'h': // Hour in day
		padstr(str, gdate.hour(), cnt);
		break;
	    case 'r': // Remove digits
		{
		    gps_type radj = t0;
		    for ( ; cnt; cnt--) radj /= 10;
		    putstr(str, radj, 0);
		}
		break;
	    case 'Y': //  Year number
		padstr(str, gdate.year(), cnt);
		break;
	    default:
		cout << "Invalid frame path: " << mLocation 
		     << " parsed as: " << str << endl;
		throw runtime_error("Invalid frame path format");
	    }
	}
    }

    //----------------------------------  Add the file name
    str += "/";
    str += mPrefix;
    str += "-";
    putstr(str, t0, 0);
    str += "-";
    gps_type delta = mDuration;
    if (limit && t0 + delta > mStop) delta = mStop - t0;
    putstr(str, delta, 0);
    str += ".gwf";
    return str;
}

//======================================  Merge frame locaters
bool
frame_loc::merge(const frame_loc& fl) {
    if (fl.mStop < mStart || fl.mStart > mStop)             return false;
    if (fl.mDuration != mDuration)                          return false;
    if (fl.mLocation != mLocation || fl.mPrefix != mPrefix) return false;
    if ((long(fl.mStart) - long(mStart)) % mDuration != 0)  return false;
    if (fl.mStop == mStart) {
	mStart = fl.mStart;
    }
    else if (fl.mStart == mStop) {
	mStop = fl.mStop;
    }
    else {
	cout << "merge: overlap " << mStart << "-" << mStop 
	     << " with: " << fl.mStart << "-" << fl.mStop << endl;
    }
    return true;
}

//======================================  Test if frame is on tape
bool
frame_loc::on_tape(gps_type t0) const {
    struct stat status;
    string fpath = file_path(t0);
    int rc = stat(fpath.c_str(), &status);
    return (!rc && status.st_blocks == 0);
}

//======================================  Test if frame is on tape
bool
frame_loc::on_tape(gps_type t0, gps_type t1) const {
    gps_type tend = (mStop < t1) ? mStop : t1;
    bool ontape = false;
    for (gps_type t=frame_start(t0); t < tend; t += mDuration) {
        if (on_tape(t)) {
	    ontape = true;
	    break;
	}
    }
    return ontape;
}

//======================================  Set location driver type
void 
frame_loc::set_driver(driver_type drv) {
    mDriver = drv;
}

//======================================  Set location string
void 
frame_loc::set_path(const std::string& loc) {
    mLocation = loc;
}

//======================================  Set data start time
void
frame_loc::set_start(gps_type t) {
    mStart = t;
}

//======================================  Set data end time
void
frame_loc::set_stop(gps_type t) {
    mStop = t;
}

//======================================  Swap contens of two frame locs
void
frame_loc::swap(frame_loc& x) {
    mLocation.swap(x.mLocation);
    mPrefix.swap(x.mPrefix);
    std::swap(mStart,    x.mStart);
    std::swap(mStop,     x.mStop);
    std::swap(mDuration, x.mDuration);
    std::swap(mDriver,   x.mDriver);
}
